from __future__ import annotations

from datetime import datetime
from typing import List, Optional

from pydantic import BaseModel, Field


class Variant(BaseModel):
    id: Optional[str]
    value: str
    stock: Optional[int]
    atc_url: Optional[str]
    direct_url: Optional[str]

    class Config:
        allow_population_by_field_name = True


class Product(BaseModel):
    monitor_id: Optional[int]
    url: str
    name: str
    brand: Optional[str]
    price: Optional[float]
    currency: Optional[str]
    sku: Optional[str]
    variants: Optional[List[Variant]] = []
    thumbnail_url: Optional[str]
    region: Optional[str]
    additional_information: Optional[str]
    found_at: datetime = Field(default_factory=datetime.utcnow)

    class Config:
        allow_population_by_field_name = True



from typing import List

import requests
from loguru import logger

from nood.api.responses import ResError, ResOK
from nood.api.types import PingProductsBody
from nood.config import API_DOMAIN
from nood.objects.product import Product


class API:
    def __init__(self, monitor_id: int, api_key: str):
        self.monitor_id = monitor_id
        self.api_key = api_key

    def ping_products(self, *, products: List[Product]) -> bool:
        if not products:
            logger.debug("No products passed to ping")
            return False
        else:
            logger.debug(f"Pinging {len(products)} products")

        # update the products site id
        for p in products:
            p.monitor_id = self.monitor_id

        # send api request
        response = requests.post(
            url=f"{API_DOMAIN}/ping/products",
            headers={
                'Content-Type': 'application/json',
                'x-api-key': self.api_key
            },
            data=PingProductsBody(products=products).json()
        )

        # handle response
        if response.status_code == 200:
            res = ResOK(**response.json())
            logger.info(f"Successfully sent {len(products)} products to "
                        f"{res.number_of_webhooks} webhooks")
            return True
        elif response.status_code == 422:
            res = ResError(**response.json())
            logger.error(f"Error sending products: {res.message}")
            return False
        else:
            logger.error(f"Unknwon response: {response.status_code} "
                         f"'{response.text}'")
            return False



from __future__ import annotations

from pydantic import BaseModel


class Response(BaseModel):
    success: bool


class ResError(Response):
    message: str


class ResOK(Response):
    monitor_id: int
    number_of_webhooks: int
    request_received_at: str
    request_fullfilled_at: str
    request_duration: float



import random
from abc import ABC, abstractmethod
from typing import List

import requests

from nood import objects
from nood.monitor.exceptions import MissingParameter


class Scraper(ABC):
    def __init__(
            self,
            url: str = None,
            pid: str = None,
            proxies: List[objects.Proxy] = None,
    ):
        if not url and not pid:
            raise MissingParameter("Task missing either 'url' or 'pid")

        self.url = url
        self.pid = pid

        if proxies is None:
            proxies = []
        self.proxies = proxies

        #
        self._static_proxy: objects.Proxy = self._set_static_proxy()
        self._s: requests.Session = self._set_session()

    def _set_static_proxy(self) -> objects.Proxy:
        self._static_proxy = self._chose_random_proxy()
        return self._static_proxy

    def _chose_random_proxy(self) -> objects.Proxy:
        if self.proxies:
            return random.choice(self.proxies)
        else:
            return objects.Proxy(host="", port="")

    def _set_session(self) -> requests.Session:
        self._s = requests.Session()
        self._set_static_proxy()
        return self._s

    def refresh_session(self):
        """Method to refresh the session object. This method also sets a
        new static proxy."""
        self._set_session()

    def get_random_proxy(self) -> objects.Proxy:
        """Get a random proxy chosen from the proxy list, returns proxy
        as dict for requests Session.
        """
        return self._chose_random_proxy().to_dict()

    def get_static_proxy(self) -> objects.Proxy:
        """Get a static proxy that stays the same for this instance, returns
        proxy as dict for requests Session.
        """
        if self._static_proxy is None:
            self._set_static_proxy()
        return self._static_proxy.to_dict()

    @abstractmethod
    def download(self) -> requests.Response:
        """This method must contain the programming logic to connect to a
        webserver, download and then return the response. It may contain logic
        to adapt links, make multiple requests, bypass protection, ..."""
        pass



from typing import Optional

from pydantic import BaseModel


class Proxy(BaseModel):
    host: str
    port: str
    user: Optional[str]
    password: Optional[str]

    def __str__(self):
        """Convert a proxy object to a string in the format of "ip:port" or
        "ip:port:user:pass".
        """

        if self.user and self.password:
            return f"{self.host}:{self.port}:{self.user}:{self.password}"
        else:
            return f"{self.host}:{self.port}"

    def to_dict(self):
        """Convert a proxy object to a dict which can be passed to the
        requests module.

        {
            'http': 'http://host:post@user:password',
            'https': 'http://host:post@user:password'
        }
        """

        if not self.host and not self.port:
            return {}

        proxy_url = f"{self.host}:{self.port}"
        if self.user and self.password:
            proxy_url = f"{self.user}:{self.password}@{proxy_url}"

        return {
            "http": f"http://{proxy_url}",
            "https": f"http://{proxy_url}",
        }

    @staticmethod
    def from_string(string: str):
        """Convert a proxy formatted as string ("ip:port" or
        "ip:port:user:pass") to a proxy object.

        :param string: "ip:port" or "ip:port:user:pass"
        :return: proxy object
        """

        host, port, *authentication = string.split(":", 4)
        if authentication:
            user, password = authentication
        else:
            user = None
            password = None

        return Proxy(host=host, port=port, user=user, password=password)

    @staticmethod
    def from_json(o: dict):
        return Proxy(**o)

    @classmethod
    def many_from_json(cls, objs):
        return [cls.from_json(o) for o in objs]
